/*
   This file is part of the index package,
   Copyright (C) 1997 Harold Tay, haroldtay@pacific.net.sg

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
   02111-1307, USA
 */

#include <stdio.h>
#include <unistd.h>
#include <fcntl.h>
#include <string.h>
#include <stdlib.h>
#include <errno.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/mman.h>
#include "index.h"

void
safewrite(void * buf, int size)
{
  int count = 0;
  while( count < size ){
    int i;
    i = write(1, buf + count, size - count);
    if( -1 == i ){
      if( EINTR == errno )continue;
      fprintf(stderr, "Can't write! %s\n", strerror(errno));
      exit(1);
    }
    count += i;
  }
}

char * map[2] = {0,};
struct meta meta[2] = {{{0,},0,},};
int fd[2] = {-1, -1};
struct stat st[2] = {{0,},};

void
cleanup(void)
{
  int i;
  for(i = 0; i < 2; i++){
    if( map[i] )munmap(map[i], st[i].st_size);
    if( fd[i] > -1 )close(fd[i]);
  }
}

int
main(int argc, char ** argv)
{
  struct meta m = {MAGIC, 0, };
  int i;

  if( argc != 3 ){
    fprintf(stderr, "Usage: %s index1 index2\n", argv[0]);
    exit(1);
  }

  atexit(cleanup);

  for(i = 0; i < 2; i++){
    if( -1 == stat(argv[i + 1], &st[i]) ){
      fprintf(stderr, "Can't stat %s: %s\n", argv[i + 1], strerror(errno));
      exit(1);
    }
    fd[i] = open(argv[i + 1], O_RDONLY);
    if( -1 == fd[i] ){
      fprintf(stderr, "Can't open %s: %s\n", argv[i + 1], strerror(errno));
      exit(1);
    }
    map[i] = mmap(0, st[i].st_size, PROT_READ, MAP_FILE|MAP_PRIVATE, fd[i], 0);
    if( (int)map[i] <= 0 ){
      fprintf(stderr, "Can't map %s: %s\n", argv[i + 1], strerror(errno));
      exit(1);
    }
    memcpy(&meta[i], map[i], sizeof(struct meta));
    if( strcmp(meta[i].magic, MAGIC) ){
      fprintf(stderr, "Bad magic in %s\n", argv[i + 1]);
      exit(1);
    }
  }

  /* Write an empty meta */
  safewrite(&meta, sizeof(meta));

  /* Write both locations sections */
  for(i = 0; i < 2; i++)
    safewrite(map[i] + sizeof(meta), meta[i].locations_size);

  /* write ONLY the first words section */
  safewrite(map[0] + sizeof(meta) + meta[i].locations_size, meta[0].words_size);

  /* The second words section will have all offsets wrong, so fix them */
  {
    char * words;
    char * end;
    int add = meta[1].locations_size;

    words = map[1] + sizeof(meta) + meta[1].locations_size;
    end = words + meta[1].words_size;

    while( words < end ){
      int * offs;
      int wlen;
      int len;
      int i;

      wlen = strlen(words);
      if( 0 != (unsigned char)*(words + wlen + 1) ){
        len = (int)*(unsigned char *)(words + wlen + 1);
        offs = (int *)(words + wlen + 1 + 1);
      }else{
        len = *(int *)(words + wlen + 1 + 1);
        offs = (int *)(words + wlen + 1 + 1 + sizeof(int));
      }

      safewrite(words, wlen + 1);
      if( len > 255 ){
        safewrite("", 1);
        safewrite(&len, sizeof(len));
      }else{
        unsigned char c = (unsigned char)len;
        safewrite(&c, 1);
      }
      for(i = 0; i < len; i++){
        int off = offs[i] + add;
        safewrite(&off, sizeof(off));
      }
      words = (char *)(offs + len);
    }
  }

  m.locations_size =
    meta[0].locations_size + meta[1].locations_size;
  m.words_size =
    meta[0].words_size + meta[1].words_size;
  if( 0 != lseek(1, 0, SEEK_SET) ){
    fprintf(stderr, "Can't seek: %s\n", strerror(errno));
    exit(1);
  }
  safewrite(&m, sizeof(meta));
  exit(0);
}





